from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple
import attr
import enchant
import enchant.checker
from enchant.tokenize import Filter
from pathlib import Path
import re
import xdg.BaseDirectory

NumberedLine = Tuple[int, str]
LinesGetter = Callable[[Path], Iterator[NumberedLine]]

SCISSORS = "# ------------------------ >8 ------------------------\n"


def get_lines_git_commit(path: Path) -> Iterator[NumberedLine]:
    with path.open() as f:
        for lineno, line in enumerate(f, start=1):
            if line == SCISSORS:
                break
            if line.startswith("#"):
                continue
            yield lineno, line


def get_lines_unknown_type(path: Path) -> Iterator[NumberedLine]:
    with path.open() as f:
        for lineno, line in enumerate(f, start=1):
            yield lineno, line


KNOWN_FILETYPES: Dict[str, LinesGetter] = {"git-commit": get_lines_git_commit}


@attr.s
class Error:
    path: Path = attr.ib()
    lineno: int = attr.ib()
    offset: int = attr.ib()
    word: str = attr.ib()
    # We need to store the whole line in order
    # to compute bytes offsets for kakoune
    line: str = attr.ib()


def get_pwl_path(lang: str) -> Path:
    data_path = Path(xdg.BaseDirectory.save_data_path("kak-spell"))
    data_path.mkdir(parents=True, exist_ok=True)
    pwl_path = data_path / f"{lang}.pwl"
    if not pwl_path.exists():
        pwl_path.write_text("")
    return pwl_path


def get_lines(path: Path, filetype: Optional[str] = None) -> Iterator[NumberedLine]:
    default_func = get_lines_unknown_type
    if filetype:
        func = KNOWN_FILETYPES.get(filetype, default_func)
    else:
        func = default_func
    yield from func(path)


class UrlFilter(Filter):  # type: ignore
    r"""Filter skipping over urls"""
    _pattern = re.compile(r"(\w+)://\S+")

    def _skip(self, word: str) -> bool:
        if self._pattern.match(word):
            return True
        return False


def get_filters() -> List[Filter]:
    return [UrlFilter]


def is_inside_backticks(text: str, error: Any) -> bool:
    pos = error.wordpos
    start = pos - 1
    end = pos + len(error.word)
    return text[start] == "`" and text[end] == "`"


class Checker:
    def __init__(self, *, lang: str):
        self.pwl_path = get_pwl_path(lang)
        dict_with_pwl = enchant.DictWithPWL(lang, str(self.pwl_path))
        self._checker = enchant.checker.SpellChecker(lang, filters=get_filters())
        self._checker.dict = dict_with_pwl

    def check(self, path: Path, filetype: Optional[str] = None) -> Iterator[Error]:
        for lineno, line in get_lines(path, filetype=filetype):
            self._checker.set_text(line)
            for error in self._checker:
                if filetype == "markdown" and is_inside_backticks(line, error):
                    continue
                yield Error(path, lineno, error.wordpos, error.word, line)

    def add(self, word: str) -> None:
        words = set(self.pwl_path.read_text().splitlines(keepends=False))
        words.add(word)
        with self.pwl_path.open("w") as f:
            for word in sorted(words):
                f.write(word + "\n")

    def remove(self, word: str) -> None:
        words = set(self.pwl_path.read_text().splitlines(keepends=False))
        words.discard(word)
        with self.pwl_path.open("w") as f:
            for word in sorted(words):
                f.write(word + "\n")

    def replace(self, word: str) -> List[str]:
        return self._checker.suggest(word)  # type: ignore



from typing import Any
from .checker import Checker  # noqa: F401
import datetime


# Useful when debugging kak scripts
def log(stuff: Any) -> None:
    with open("/tmp/kak-spell.log", "a") as f:
        f.write(str(datetime.datetime.now()) + " " + str(stuff) + "\n")



from typing import List, Tuple
from . import log
from .checker import Error


def handle_spelling_errors(errors: List[Error], *, timestamp: int, lang: str) -> None:
    insert_errors_in_scratch_buffer(errors, lang=lang)
    set_spell_errors(errors, timestamp=timestamp)


def show_spelling_buffer() -> None:
    print("edit -existing *spelling*")


def insert_errors_in_scratch_buffer(errors: List[Error], lang: str) -> None:
    # TODO: -fifo ?
    print("edit -scratch *spelling*")
    cmd = "execute-keys \\% <ret> d i %{"
    for error in errors:
        line, col, length = error_to_range(error)
        cmd += f"{line}.{col},{line}.{col+length-1} {error.word}<ret>"
    cmd += "}"
    cmd += "<esc> gg"
    cmd += "\n" "map buffer normal '<ret>' ':<space>kak-spell-jump<ret>'"
    cmd += (
        "\n"
        f"map buffer normal 'a' ':<space>kak-spell-add-from-spelling-buffer {lang}<ret>'"
    )
    cmd += "\n" "execute-keys <esc> ga"
    log(cmd)
    print(cmd)


def set_spell_errors(errors: List[Error], *, timestamp: int) -> None:
    n = 0
    cmd = f"set-option buffer spell_errors {timestamp}"
    for error in errors:
        (line, col, length) = error_to_range(error)
        cmd += f" {line}.{col}+{length}|Error"
        n += 1
    log(cmd)
    print(cmd)
    if n:
        print("echo -markup {red}", n, "spelling error(s)")
    else:
        show_no_errors()


def menu_from_replacements(replacements: List[str]) -> None:
    if not replacements:
        print("fail no suggestions")
        return
    menu = ""
    for entry in replacements:
        # Note: %{...} is kakoune way of grouping stuff that may - or not
        # contains quotes, which prevents us from using `%`, `.format()` or
        # f-strings :P
        # also note that we call 'write' and 'kak-spell' *in* the menu action, otherwise
        # if you call `write` *after* the menu command, the buffer is saved without the new
        # contents written !
        menu_entry = "%{ENTRY} %{execute-keys -itersel %{cENTRY<esc>be} "
        menu_entry += ":write <ret> :kak-spell <ret>}"
        menu_entry = menu_entry.replace("ENTRY", entry)
        menu += " " + menu_entry
    cmd = "menu " + menu
    print(cmd)


def show_no_errors() -> None:
    print("echo -markup {green} no spelling errors")


def goto_next(pos_option: str, ranges_option: str) -> None:
    cursor = convert_to_pos(pos_option)
    ranges = convert_ranges(ranges_option)
    if not ranges:
        show_no_errors()
        return
    next_selection = get_next_selection(cursor, ranges)
    goto_and_select(next_selection)


def goto_previous(pos_option: str, ranges_option: str) -> None:
    cursor = convert_to_pos(pos_option)
    ranges = convert_ranges(ranges_option)
    if not ranges:
        show_no_errors()
        return
    previous_selection = get_previous_selection(cursor, ranges)
    goto_and_select(previous_selection)


# (line, col)
Pos = Tuple[int, int]

# (line, col, length)
Range = Tuple[int, int, int]


def error_to_range(error: Error) -> Range:
    bytes_offset = len((error.line[: error.offset]).encode())
    return (error.lineno, bytes_offset + 1, len(error.word.encode()))


def convert_to_pos(pos_option: str) -> Pos:
    start, end = pos_option.split(".")
    return int(start), int(end)


def convert_ranges(ranges_option: str) -> List[Range]:
    if ranges_option == "0":
        return []

    res = []
    ranges_list = ranges_option.split()[1:]
    for range_option in ranges_list:
        range_str = range_option.split("|")[0]  # strip face
        start, end = range_str.split(",")
        start_line, start_col = convert_to_pos(start)
        end_line, end_col = convert_to_pos(end)
        res.append((start_line, start_col, end_col))
    return res


def goto_and_select(range: Range) -> None:
    line, col, end = range
    length = end - col
    cmd = f"select {line}.{col},{line}.{col+length}"
    print(cmd)


def get_next_selection(cursor: Pos, ranges: List[Range]) -> Range:
    cursor_line, cursor_col = cursor
    for r in ranges:
        start_line, start_col, end_col = r
        if start_line < cursor_line:
            continue
        if start_line == cursor_line and end_col <= cursor_col:
            continue
        return r
    # if we reach there, return the first error (auto-wrap)
    return ranges[0]


def get_previous_selection(cursor: Pos, ranges: List[Range]) -> Range:
    cursor_line, cursor_col = cursor
    for r in reversed(ranges):
        start_line, start_col, end_col = r
        if start_line > cursor_line:
            continue
        if start_line == cursor_line and end_col > cursor_col:
            continue
        return r
    # if we reach there, return the last error (auto-wrap)
    return ranges[-1]



#!/usr/bin/env python3
from typing import List, Optional
import argparse
import sys

from pathlib import Path

from .checker import Checker
from . import kak
from . import log


def add_word(word: str, *, verbose: bool = True, lang: str) -> None:
    checker = Checker(lang=lang)
    checker.add(word)
    if verbose:
        print("Word added to", checker.pwl_path)


def remove_word(word: str, *, verbose: bool = True, lang: str) -> None:
    checker = Checker(lang=lang)
    checker.remove(word)
    if verbose:
        print("Word removed from", checker.pwl_path)


def check(
    path: Path,
    *,
    lang: str,
    filetype: Optional[str] = None,
    kakoune: bool = False,
    kak_timestamp: int = 0,
) -> bool:
    checker = Checker(lang=lang)
    errors = checker.check(path, filetype=filetype)
    if kakoune:
        # We need a real list because we'll iterate on it twice
        kak.handle_spelling_errors(list(errors), timestamp=kak_timestamp, lang=lang)
        return True
    else:
        ok = True
        for error in errors:
            ok = False
            print(f"{path}:{error.lineno}:{error.offset + 1}: error: {error.word}")
    return ok


def replace(word: str, *, lang: str, kak_output: bool) -> None:
    checker = Checker(lang=lang)
    replacements = checker.replace(word)
    if kak_output:
        kak.menu_from_replacements(replacements)
    else:
        print(" ".join(replacements))


def main(argv: Optional[List[str]] = None) -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("--lang", default="en_US")
    parser.add_argument("--quiet", action="store_false", dest="verbose")
    parser.set_defaults(verbose=True)

    subparsers = parser.add_subparsers(title="commands", dest="command")

    check_parser = subparsers.add_parser("check")
    check_parser.add_argument("path", type=Path)
    check_parser.add_argument(
        "--filetype", help="file type, as set by kakoune in $kak_opt_filetype"
    )
    check_parser.add_argument(
        "--kakoune", help="output kakoune commands", action="store_true"
    )
    check_parser.add_argument("--kak-timestamp", help="kakoune timestamp", type=int)

    list_parser = subparsers.add_parser("list")
    list_parser.add_argument("path", type=Path)
    list_parser.add_argument(
        "--filetype", help="file type, as set by kakoune in $kak_opt_filetype"
    )

    add_parser = subparsers.add_parser("add")
    add_parser.add_argument("word")

    remove_parser = subparsers.add_parser("remove")
    remove_parser.add_argument("word")

    replace_parser = subparsers.add_parser("replace")
    replace_parser.add_argument(
        "--kakoune", action="store_true", help="Output kak-script compatible output"
    )
    replace_parser.add_argument("word")

    next_parser = subparsers.add_parser("next")
    next_parser.add_argument("--ranges", help="kakoune ranges", required=True)
    next_parser.add_argument("--pos", help="kakoune cursor pos", required=True)

    previous_parser = subparsers.add_parser("previous")
    previous_parser.add_argument("--ranges", help="kakoune ranges", required=True)
    previous_parser.add_argument("--pos", help="kakoune cursor pos", required=True)

    args = parser.parse_args(args=argv)
    log(args)
    lang = args.lang
    verbose = args.verbose

    if args.command == "add":
        word = args.word
        add_word(word, verbose=verbose, lang=lang)
    elif args.command == "remove":
        word = args.word
        remove_word(word, verbose=verbose, lang=lang)
    elif args.command == "check":
        ok = check(
            args.path,
            lang=lang,
            filetype=args.filetype,
            kakoune=args.kakoune,
            kak_timestamp=args.kak_timestamp,
        )
        if not ok:
            sys.exit(1)
    elif args.command == "list":
        kak.show_spelling_buffer()
    elif args.command == "replace":
        word = args.word
        kakoune = args.kakoune
        replace(word, lang=lang, kak_output=kakoune)
    elif args.command == "next":
        kak.goto_next(args.pos, args.ranges)
    elif args.command == "previous":
        kak.goto_previous(args.pos, args.ranges)
    else:
        parser.print_help()
        sys.exit(1)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kak_spell']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=19.3.0,<20.0.0', 'pyenchant==3.0.1', 'pyxdg>=0.26,<0.27']

entry_points = \
{'console_scripts': ['kak-spell = kak_spell.cli:main']}

setup_kwargs = {
    'name': 'kak-spell',
    'version': '0.3.1',
    'description': 'PyEnchant wrapper for Kakoune',
    'long_description': '# kak-spell\n\nPyEnchant wrapper for Kakoune.\n\n## Installation\n\n\n1. Install the C Enchant library, and the required dictionaries. See [PyEnchant documentation](https://pyenchant.github.io/pyenchant/install.html) for details.\n\n2. Install the `kak-spell` script, for instance with [pipx](https://pipxproject.github.io/pipx/):\n\n```\npipx install kak-spell\n```\n\n3. Install [plug.kak](https://github.com/andreyorst/plug.kak) and add the following lines in your `kakrc`:\n\n```kak\nplug "dmerejkowsky/kak-spell"\n```\n\n\n4. (optional): declare a user mode and some mappings:\n\n```kak\nplug "dmerejkowsky/kak-spell" config %{\n  declare-user-mode kak-spell\n  map global user s \': enter-user-mode -lock kak-spell<ret>\' -docstring \'enter spell user mode\'\n  map global kak-spell a \': kak-spell-add<ret>\' -docstring \'add the selection to the user dict\'\n  map global kak-spell d \': kak-spell-disable<ret>\' -docstring \'clear spelling highlighters\'\n  map global kak-spell e \': kak-spell-enable en_US<ret> :kak-spell <ret>\' -docstring \'enable spell check in English\'\n  map global kak-spell f \': kak-spell-enable fr_FR<ret> :kak-spell <ret>\' -docstring \'run spell check in French\'\n  map global kak-spell l \': kak-spell-list <ret>\' -docstring \'list spelling errors in a buffer\'\n  map global kak-spell n \': kak-spell-next<ret>\' -docstring \'go to next spell error\'\n  map global kak-spell p \': kak-spell-previous<ret>\' -docstring \'go to next spell error\'\n  map global kak-spell r \': kak-spell-replace<ret>\' -docstring \'suggest a list of replacements\'\n  map global kak-spell x \': kak-spell-remove<ret>\' -docstring \'remove the selection from the user dict\'\n}\n```\n\nNote that `kak-spell-enable` does several things:\n* Set a buffer scoped option `kak_spell_lang` that is used by other commands\n* Add a highlighter so that spelling errors are colored with the `Error` face\n* Adds a `BufWritePost` hook to spell check the buffer each time it gets written\n\nThe command `kak-spell-disable` undoes all of the above.\n\nFor now, there\'s no option to disable the hook, or to have it run in response to other events. Please open an issue if this bothers you.\n\n## Discuss\n\nYou can discuss features of this plugin on [discuss.kakoune.com](https://discuss.kakoune.com/t/alternate-implementation-for-spell-checker/781).\n\nI\'d like to thank all the people who contributed code and ideas to make usage of this plugin easier.\n',
    'author': 'Dimitri Merejkowsky',
    'author_email': 'd.merej@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/dmerejkowsky/kak-spell',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
